# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

from ._string_widget import MultiStringWidget, StringWidget


class FilenameWidget(StringWidget): ...


class MultiFilenameWidget(MultiStringWidget): ...
